////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FFGL.cpp
//
// FreeFrame is an open-source cross-platform real-time video effects plugin system.
// It provides a framework for developing video effects plugins and hosts on Windows,
// Linux and Mac OSX.
//
// FreeFrameGL (FFGL) is an extension to the FreeFrame spec to support video processing
// with OpenGL on Windows, Linux, and Mac OSX.
//
// Copyright (c) 2018 www.freeframe.org
// All rights reserved.
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Redistribution and use in source and binary forms, with or without modification,
//	are permitted provided that the following conditions are met:
//
//  * Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
//  * Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in
//    the documentation and/or other materials provided with the
//    distribution.
//  * Neither the name of FreeFrame nor the names of its
//    contributors may be used to endorse or promote products derived
//    from this software without specific prior written permission.
//
//
//	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
//	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
//	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
//	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
//	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
//	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
//	OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
//	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
//	OF THE POSSIBILITY OF SUCH DAMAGE.
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// First version, Marcus Clements (marcus@freeframe.org)
// www.freeframe.org
//
// FreeFrame 1.0 upgrade by Russell Blakeborough
// email: boblists@brightonart.org
//
// FreeFrame 1.0 - 03 upgrade
// and implementation of FreeFrame SDK methods by Gualtiero Volpe
// email: Gualtiero.Volpe@poste.it
//
// FFGL upgrade by Trey Harrison
// email: trey@harrisondigitalmedia.com
//
// FFGL 2.0 by Menno Vink (menno@resolume.com)
// www.resolume.com
// -Removed the old FreeFrame functionality, and thus CPU processing is no longer supported.
// -Completed the FF_TYPE_OPTION parameter type. Plugins can add parameter elements which hosts
//  should use to show a dropdown styled parameter. When one of the options is selected the parameter
//  will get a value equal to that of the option's value.
// -Completed the FF_TYPE_BUFFER parameter type. When this parameter is used it should provide a usage indication
//  which the host can then use to fill that buffer with the requested data. This can be used by plugins to
//  access the host's global fft data for example.
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "FFGLPluginSDK.h"
#include <memory.h>
#include "../glsdk_0_5_2/glload/include/gl_load.h"

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Static and extern variables used in the FreeFrame SDK
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

extern CFFGLPluginInfo* g_CurrPluginInfo;

static CFreeFrameGLPlugin* s_pPrototype = NULL;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FreeFrame SDK default implementation of the FreeFrame global functions.
// Such function are called by the plugMain function, the only function a plugin exposes.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool InitGLExts()
{
#if defined( WIN32 )
	static bool triedInit  = false;
	static bool initResult = false;
	if( triedInit )
		return initResult;
	triedInit = true;

	if( ogl_LoadFunctions() == ogl_LOAD_FAILED )
		return false;
	initResult = true;
	return initResult;
#else
	return true;
#endif
}

// 2019-07-11 ffgl@subpixels.com - changed return type from (void*) to (const PluginInfoStruct*)
const PluginInfoStruct* getInfo()
{
	return g_CurrPluginInfo->GetPluginInfo();
}
FFResult initialise()
{
	if( g_CurrPluginInfo == NULL )
		return FF_FAIL;

	if( s_pPrototype == NULL )
	{
		//get the instantiate function pointer
		FPCREATEINSTANCEGL* pInstantiate = g_CurrPluginInfo->GetFactoryMethod();

		//call the instantiate function
		FFResult ret = pInstantiate( &s_pPrototype );

		//make sure the instantiate call worked
		if( ( ret == FF_FAIL ) || ( s_pPrototype == NULL ) )
			return FF_FAIL;

		return FF_SUCCESS;
	}

	return FF_SUCCESS;
}

FFResult deInitialise()
{
	if( s_pPrototype != NULL )
	{
		delete s_pPrototype;
		s_pPrototype = NULL;
	}
	return FF_SUCCESS;
}

unsigned int getNumParameters()
{
	if( s_pPrototype == NULL )
	{
		FFResult dwRet = initialise();
		if( dwRet == FF_FAIL )
			return FF_FAIL;
	}

	return s_pPrototype->GetNumParams();
}

// 2019-07-11 ffgl@subpixels.com - changed return type from (char*) to (const char*)
const char* getParameterName( unsigned int index )
{
	if( s_pPrototype == NULL )
	{
		FFResult dwRet = initialise();
		if( dwRet == FF_FAIL )
			return NULL;
	}

	return s_pPrototype->GetParamName( index );
}
FFMixed getParameterDefault( unsigned int index )
{
	FFMixed ret;
	ret.UIntValue = FF_FAIL;
	if( s_pPrototype == NULL )
	{
		FFResult dwRet = initialise();
		if( dwRet == FF_FAIL )
			return ret;
	}
	return s_pPrototype->GetParamDefault( index );
}

//2019-07-27 ffgl@subpixels.com - Use enum FFPluginCapability, rename index to capability
FFResult getPluginCaps( FFPluginCapability capability )
{
	int MinInputs = -1;
	int MaxInputs = -1;

	if( s_pPrototype == NULL )
	{
		FFResult dwRet = initialise();
		if( dwRet == FF_FAIL )
			return FF_FAIL;
	}

	switch( capability )
	{
	case FFPluginCapability::FF_CAP_SETTIME:
		if( s_pPrototype->GetTimeSupported() )
			return FF_TRUE;
		else
			return FF_FALSE;
	case FFPluginCapability::FF_CAP_MINIMUMINPUTFRAMES:
		MinInputs = s_pPrototype->GetMinInputs();
		if( MinInputs < 0 )
			return FF_FALSE;
		return MinInputs;
	case FFPluginCapability::FF_CAP_MAXIMUMINPUTFRAMES:
		MaxInputs = s_pPrototype->GetMaxInputs();
		if( MaxInputs < 0 )
			return FF_FALSE;
		return MaxInputs;
	}

	return FF_FALSE;
}

// 2019-07-11 ffgl@subpixels.com - changed return type from (void*) to (PluginExtendedInfoStruct*)
const PluginExtendedInfoStruct* getExtendedInfo()
{
	return g_CurrPluginInfo->GetPluginExtendedInfo();
}

FFParameterType getParameterType( unsigned int index )
{
	if( s_pPrototype == NULL )
	{
		FFResult dwRet = initialise();
		if( dwRet == FF_FAIL )
			return FFParameterType::FF_FAIL;
	}

	return s_pPrototype->GetParamType( index );
}

// 2019-07-11 ffgl@subpixels.com
// - changed return type from (void*) to (CFreeFrameGLPlugin*)
// - use const pFF_FAIL to return correct type
// - use FFMixed::StrValue and FloatValue instead of pointer recasting
// 2019-07-22 ffgl@subpixels.com
// - use enum FFParameterType
CFreeFrameGLPlugin* instantiateGL( const FFGLViewportStruct* pGLViewport )
{
	// Convenience constant to return on failure.
	CFreeFrameGLPlugin* const pFF_FAIL = (CFreeFrameGLPlugin*)FF_FAIL;

	if( g_CurrPluginInfo == NULL || pGLViewport == NULL )
		return pFF_FAIL;

	// If the plugin is not initialized, initialize it
	if( s_pPrototype == NULL )
	{
		FFResult dwRet = initialise();

		if( ( dwRet == FF_FAIL ) || ( s_pPrototype == NULL ) )
			return pFF_FAIL;
	}

	//get the instantiate function pointer
	FPCREATEINSTANCEGL* pInstantiate = g_CurrPluginInfo->GetFactoryMethod();

	CFreeFrameGLPlugin* pInstance = NULL;

	//call the instantiate function
	FFResult dwRet = pInstantiate( &pInstance );

	//make sure the instantiate call worked
	if( ( dwRet == FF_FAIL ) || ( pInstance == NULL ) )
		return pFF_FAIL;

	pInstance->m_pPlugin = pInstance;

	// Initializing instance with default values
	for( unsigned int i = 0; i < s_pPrototype->GetNumParams(); ++i )
	{
		FFParameterType pType = (FFParameterType) s_pPrototype->GetParamType( i );
		FFMixed pDefault   = s_pPrototype->GetParamDefault( i );
		if( pType == FFParameterType::FF_TYPE_TEXT )
			dwRet = pInstance->SetTextParameter( i, pDefault.StrValue );
		else
			dwRet = pInstance->SetFloatParameter( i, pDefault.FloatValue );

		//#ifdef FFGLTEXTFIX

		//    int type = s_pPrototype->GetParamType(i);
		//
		//    switch( type )
		//    {
		//      case FFParameterType::FF_TYPE_TEXT:
		//		SetParameterStruct ParamStruct;
		//		ParamStruct.NewParameterValue = pDefault;
		//        dwRet = pInstance->SetParameter(&ParamStruct);
		//        break;
		//      case FFParameterType::FFParameterType::FF_TYPE_BUFFER:
		//        {
		//          /*int n = s_pPrototype->GetNumParamElements(DWORD(i));
		//          float * buf = new float[n];
		//          for( int i = 0; i < n; i++ )
		//            buf[i] = 0.0f; // TODO: use parameter default?
		//          ParamStruct.NewParameterValue = (DWORD)buf;
		//          dwRet = pInstance->SetParameter(&ParamStruct);
		//          delete [] buf;*/
		//        }
		//        break;
		//		default:
		//         memcpy(&ParamStruct.NewParameterValue, pValue, 4);
		//         dwRet = pInstance->SetParameter(&ParamStruct);
		//         break;
		//
		//    }

		if( dwRet == FF_FAIL )
		{
			//SetParameter failed, delete the instance
			delete pInstance;
			return pFF_FAIL;
		}
	}

	if( !InitGLExts() )
		return pFF_FAIL;

	//call the InitGL method
	if( pInstance->InitGL( pGLViewport ) == FF_SUCCESS )
	{
		//succes? we're done.
		return pInstance;
	}

	//InitGL failed, delete the instance
	pInstance->DeInitGL();
	delete pInstance;

	return pFF_FAIL;
}

FFResult deInstantiateGL( void* instanceID )
{
	CFreeFrameGLPlugin* p = (CFreeFrameGLPlugin*)instanceID;

	if( p != NULL )
	{
		// Disconnect if necessary
		if( p->m_isConnected )
		{
			p->Disconnect();
			p->m_isConnected = false;
		}

		p->DeInitGL();
		delete p;

		return FF_SUCCESS;
	}

	return FF_FAIL;
}

FFUInt32 getNumParameterElements( unsigned int index )
{
	if( s_pPrototype == NULL )
	{
		FFResult dwRet = initialise();
		if( dwRet == FF_FAIL )
			return FF_FAIL;
	}

	return s_pPrototype->GetNumParamElements( index );
}

// 2019-07-11 ffgl@subpixels.com - changed return type from (char*) to (const char*)
const char* getParameterElementName( unsigned int paramIndex, unsigned int elementIndex )
{
	if( s_pPrototype == NULL )
	{
		FFResult dwRet = initialise();
		if( dwRet == FF_FAIL )
			return NULL;
	}

	return s_pPrototype->GetParamElementName( paramIndex, elementIndex );
}

FFMixed getParameterElementDefault( unsigned int paramIndex, unsigned int elementIndex )
{
	FFMixed ret;
	ret.UIntValue = FF_FAIL;
	if( s_pPrototype == NULL )
	{
		FFResult dwRet = initialise();
		if( dwRet == FF_FAIL )
			return ret;
	}
	return s_pPrototype->GetParamElementDefault( paramIndex, elementIndex );
}

// 2019-07-22 ffgl@subpixels.com - use enum FFParameterUsage
FFParameterUsage getParameterUsage( unsigned int index )
{
	if( s_pPrototype == NULL )
	{
		FFResult dwRet = initialise();
		if( dwRet == FF_FAIL )
			return FFParameterUsage::FF_FAIL;
	}

	return s_pPrototype->GetParamUsage( index );
}

const char* getPluginShortName()
{
	if( s_pPrototype == NULL )
	{
		FFResult dwRet = initialise();
		if( dwRet == FF_FAIL )
			return NULL;
	}

	const char* shortName = s_pPrototype->GetShortName();
	if( shortName == NULL )
		return NULL;

	return shortName;
}

// 2019-07-11 ffgl@subpixels.com
// - use FFMixed::pGetRange instead of pointer recasting
FFMixed getParamRange( FFMixed input )
{
	FFMixed ret;
	ret.UIntValue = FF_FAIL;
	if( s_pPrototype == NULL )
	{
		FFResult dwRet = initialise();
		if( dwRet == FF_FAIL )
			return ret;
	}
	ret.UIntValue = FF_SUCCESS;

	GetRangeStruct* getRange = input.pGetRange;

	RangeStruct range = s_pPrototype->GetParamRange( getRange->parameterNumber );
	getRange->range   = range;
	return ret;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Implementation of plugMain, the one and only exposed function
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// 2019-07-11 ffgl@subpixels.com
// - use new FFMixed union value types and pointers instead of pointer recasting
// 2019-07-22 ffgl@subpixels.com
// - use enums FFParameterType, FFPluginCapability

#ifdef WIN32

FFMixed __stdcall plugMain( FFUInt32 functionCode, FFMixed inputValue, FFInstanceID instanceID )

#elif TARGET_OS_MAC

FFMixed plugMain( FFUInt32 functionCode, FFMixed inputValue, FFInstanceID instanceID )

#elif __linux__

FFMixed plugMain( FFUInt32 functionCode, FFMixed inputValue, FFInstanceID instanceID )

#endif

{
#if defined( _DEBUG ) && defined(WIN32)
	LaunchDebugger();
#endif

	FFMixed retval;
	retval.UIntValue = FF_FAIL;// Assume failure until we succeed

	// declare pPlugObj - pointer to this instance
	CFreeFrameGLPlugin* pPlugObj;

	// typecast DWORD into pointer to a CFreeFrameGLPlugin
	pPlugObj = (CFreeFrameGLPlugin*)instanceID;

	switch( (FFFunctionCode) functionCode )
	{
	case FFFunctionCode::FF_GETINFO:
		retval.pPluinInfo = getInfo();
		break;
	case FFFunctionCode::FF_INITIALISE_V2:
		retval.UIntValue = initialise();
		break;
	case FFFunctionCode::FF_DEINITIALISE:
		retval.UIntValue = deInitialise();
		break;
	case FFFunctionCode::FF_GETNUMPARAMETERS:
		retval.UIntValue = getNumParameters();
		break;
	case FFFunctionCode::FF_GETPARAMETERNAME:
		retval.StrValue = getParameterName( inputValue.UIntValue );
		break;
	case FFFunctionCode::FF_GETPARAMETERDEFAULT:
		retval = getParameterDefault( inputValue.UIntValue );
		break;
	case FFFunctionCode::FF_GETPARAMETERDISPLAY:
		if( pPlugObj != NULL )
			retval.StrValue = pPlugObj->GetParameterDisplay( inputValue.UIntValue );
		else
			retval.StrValue = (char*)FF_FAIL;
		break;
	case FFFunctionCode::FF_SETPARAMETER:
		if( pPlugObj != NULL )
		{
			const auto args = inputValue.pSetParameter;

			if( getParameterType( args->ParameterNumber ) == FFParameterType::FF_TYPE_TEXT )
				retval.UIntValue = pPlugObj->SetTextParameter( args->ParameterNumber, args->NewParameterValue.StrValue );
			else
				retval.UIntValue = pPlugObj->SetFloatParameter( args->ParameterNumber, args->NewParameterValue.FloatValue );
		}
		break;
	case FFFunctionCode::FF_GETPARAMETER:
		if( pPlugObj != NULL )
		{
			if( getParameterType( inputValue.UIntValue ) == FFParameterType::FF_TYPE_TEXT )
				retval.StrValue = pPlugObj->GetTextParameter( inputValue.UIntValue );
			else
				retval.FloatValue = pPlugObj->GetFloatParameter( inputValue.UIntValue );
		}
		break;
	case FFFunctionCode::FF_GETPLUGINCAPS:
		retval.UIntValue = getPluginCaps( (FFPluginCapability) inputValue.UIntValue );
		break;
	case FFFunctionCode::FF_GETEXTENDEDINFO:
		retval.pPluginExtendedInfo = getExtendedInfo();
		break;
	case FFFunctionCode::FF_GETPARAMETERTYPE:
		retval.ParameterType = getParameterType( inputValue.UIntValue );
		break;
	case FFFunctionCode::FF_GETINPUTSTATUS:
		if( pPlugObj != NULL )
			retval.InputStatus = pPlugObj->GetInputStatus( inputValue.UIntValue );
		break;
	case FFFunctionCode::FF_PROCESSOPENGL:
		if( pPlugObj != NULL && inputValue.pProcessOpenGL != NULL )
		{
			// Ensure Connect has been called
			if( !pPlugObj->m_isConnected )
			{
				pPlugObj->Connect();
				pPlugObj->m_isConnected = true;
			}

			retval.UIntValue = pPlugObj->ProcessOpenGL( inputValue.pProcessOpenGL );
		}
		break;
	case FFFunctionCode::FF_INSTANTIATEGL:
		retval.pCFreeFrameGLPlugin = instantiateGL( inputValue.pFFGLViewport );
		break;
	case FFFunctionCode::FF_DEINSTANTIATEGL:
		if( pPlugObj != NULL )
			retval.UIntValue = deInstantiateGL( pPlugObj );
		break;
	case FFFunctionCode::FF_SETTIME:
		if( pPlugObj != NULL && inputValue.pDouble != NULL )
			retval.UIntValue = pPlugObj->SetTime( *inputValue.pDouble );
		break;
	case FFFunctionCode::FF_CONNECT:
		if( pPlugObj != NULL )
		{
			retval.UIntValue = pPlugObj->Connect();
			if( retval.UIntValue == FF_SUCCESS )
				pPlugObj->m_isConnected = true;
		}
		break;
	case FFFunctionCode::FF_DISCONNECT:
		if( pPlugObj != NULL )
		{
			retval.UIntValue = pPlugObj->Disconnect();
			if( retval.UIntValue == FF_SUCCESS )
				pPlugObj->m_isConnected = false;
		}
		break;
	case FFFunctionCode::FF_RESIZE:
		if( pPlugObj != NULL )
			retval.UIntValue = pPlugObj->Resize( inputValue.pFFGLViewport );
		break;
	case FFFunctionCode::FF_GETNUMPARAMETERELEMENTS:
		retval.UIntValue = getNumParameterElements( inputValue.UIntValue );
		break;
	case FFFunctionCode::FF_GET_PARAMETER_ELEMENT_NAME:
	{
		const auto args = inputValue.pGetParameterElementName;
		retval.StrValue = getParameterElementName( args->ParameterNumber, args->ElementNumber );
		break;
	}
	case FFFunctionCode::FF_GET_PARAMETER_ELEMENT_DEFAULT:
	{
		const auto args = inputValue.pGetParameterElementValue;
		retval          = getParameterElementDefault( args->ParameterNumber, args->ElementNumber );
		break;
	}
	case FFFunctionCode::FF_SET_PARAMETER_ELEMENT_VALUE:
		if( pPlugObj != NULL )
		{
			const auto args = inputValue.pSetParameterElementValue;
			retval.UIntValue = pPlugObj->SetParamElementValue( args->ParameterNumber, args->ElementNumber, args->NewParameterValue.FloatValue );
		}
		break;
	case FFFunctionCode::FF_GETPARAMETERUSAGE:
		retval.ParameterUsage = getParameterUsage( inputValue.UIntValue );
		break;
	case FFFunctionCode::FF_GETPLUGINSHORTNAME:
		retval.StrValue = getPluginShortName();
		break;
	case FFFunctionCode::FF_SET_BEATINFO:
		if( pPlugObj != NULL )
		{
			const auto args = inputValue.pSetBeatinfo;
			pPlugObj->SetBeatInfo( args->bpm, args->barPhase );
			retval.UIntValue = FF_SUCCESS;
		}
		break;
	case FFFunctionCode::FF_SET_HOSTINFO:
		if( pPlugObj != NULL )
		{
			const auto args = inputValue.pSetHostinfo;
			pPlugObj->SetHostInfo( args->name, args->version );
			retval.UIntValue = FF_SUCCESS;
		}
		break;
	case FFFunctionCode::FF_SET_SAMPLERATE:
		if( pPlugObj != NULL )
		{
			pPlugObj->SetSampleRate( inputValue.UIntValue );
			retval.UIntValue = FF_SUCCESS;
		}
		break;
	case FFFunctionCode::FF_GET_RANGE:
		//TODO: 2019-07-17 ffgl@subpixels.com - investigate whether inputValue should be used to pass back the range
		retval = getParamRange( inputValue );
		break;

		//Previously used function codes that are no longer supported:
		//case FFFunctionCode::FF_INITIALISE:
		/**
		 * We're dropping the old FFGL 1.6 and lower initialise here. FFGL 2.0 removed old stuff and made support for newer stuff mandatory
		 * so hosts need a way to know they cannot use this plugin if they're dependant on the old behaviour. If the host isn't dependant on the old
		 * behaviour it will have to update to build using the FFGL 2.0 sdk and instead invoke the initialise_v2 opcode. This way
		 * the plugin and host both agree that it's okay not to support the old behaviour.
		 */
		//case FFFunctionCode::FF_INSTANTIATE:
		//case FFFunctionCode::FF_DEINSTANTIATE:
		//case FFFunctionCode::FF_PROCESSFRAME:
		//case FFFunctionCode::FF_PROCESSFRAMECOPY:
	}

	return retval;
}
