#include "AddSubtract.h"
using namespace ffglex;

static CFFGLPluginInfo PluginInfo(
	PluginFactory< AddSubtract >,// Create method
#ifdef _DEBUG
	"?e01",                      // Plugin unique ID of maximum length 4.
	"?SubAdd",                   // Plugin name
#else
	"_e01",                      // Plugin unique ID of maximum length 4.
	"_SubAdd",                   // Plugin name
#endif
	2,                           // API major version number
	1,                           // API minor version number
	1,                           // Plugin major version number
	0,                           // Plugin minor version number
	//FF_EFFECT,                   // Plugin type
	FFPluginType::FF_EFFECT,     // Plugin type
	"Subtract and Add colours",  // Plugin description
	"Resolume FFGL Example (mod by ffgl@subpixels.com)" // About
);

static const char _vertexShaderCode[] = R"(#version 410 core
uniform vec2 MaxUV;

layout( location = 0 ) in vec4 vPosition;
layout( location = 1 ) in vec2 vUV;

out vec2 uv;

void main()
{
	gl_Position = vPosition;
	uv = vUV * MaxUV;
}
)";

static const char _fragmentShaderCode[] = R"(#version 410 core
uniform sampler2D InputTexture;
uniform vec3 Brightness;

in vec2 uv;

out vec4 fragColor;

void main()
{
	vec4 color = texture( InputTexture, uv );
	//The InputTexture contains premultiplied colors, so we need to unpremultiply first to apply our effect on straight colors.
	if( color.a > 0.0 )
		color.rgb /= color.a;

	color.rgb += Brightness * 2. - 1.;

	//The plugin has to output premultiplied colors, this is how we're premultiplying our straight color while also
	//ensuring we aren't going out of the LDR the video engine is working in.
	color.rgb = clamp( color.rgb * color.a, vec3( 0.0 ), vec3( color.a ) );
	fragColor = color;
}
)";

AddSubtract::AddSubtract()
{
	// Input properties
	SetMinInputs( 1 );
	SetMaxInputs( 1 );

	//We declare that this plugin has a Brightness parameter which is a RGB param.
	//The name here must match the one you declared in your fragment shader.
	addRGBColorParam( "Brightness" );

	ParamOption::Options options;
	options.push_back( { "Test1", 1.0 } );
	options.push_back( { "Test2", 2.0 } );
	options.push_back( { "Test3", 3.0 } );

	//// Test Parameters
	//addParam( ParamBool::create( "TestBool" ) );
	//addParam( ParamEvent::create( "TestEvent" ) );
	//addParam( ParamFFT::create( "TestFFT" ) );
	//addParam( ParamOption::create( "TestOption", options ) );
	//addParam( ParamRange::create( "TestRange", 15.0f, { 0.0f, 100.0f } ) );
	//addParam( ParamText ::create( "TestText" ) );
	//addParam( ParamTrigger ::create( "TestTrigger" ) );

	// Test Parameters
	addParam( new ParamBool( "TestBool" ) );
	addParam( new ParamEvent( "TestEvent" ) );
	addParam( new ParamFFT( "TestFFT", 8 ) );
	addParam( new ParamOption( "TestOption", options ) );
	addParam( new ParamRange( "TestRange", 15.0f, { 0.0f, 100.0f } ) );
	addParam( new ParamText( "TestText" ) );
	addParam( new ParamTrigger( "TestTrigger" ) );

	SetTimeSupported( true );
}
AddSubtract::~AddSubtract()
{
}

FFResult AddSubtract::InitGL( const FFGLViewportStruct* vp )
{
	if( !shader.Compile( _vertexShaderCode, _fragmentShaderCode ) )
	{
		DeInitGL();
		return FF_FAIL;
	}
	if( !quad.Initialise() )
	{
		DeInitGL();
		return FF_FAIL;
	}

	//FFGL requires us to leave the context in a default state on return
	resetOpenGLState();
	//Use base-class init as success result so that it retains the viewport.
	return CFreeFrameGLPlugin::InitGL( vp );
}
FFResult AddSubtract::ProcessOpenGL( ProcessOpenGLStruct* pGL )
{
	if( pGL->numInputTextures < 1 )
		return FF_FAIL;

	if( pGL->inputTextures[ 0 ] == NULL )
		return FF_FAIL;

	updateAudioAndTime();

	shader.Use();
	shader.Bind( "inputTexture", 0, *pGL->inputTextures[ 0 ] );

	//The input texture's dimension might change each frame and so might the content area.
	//We're adopting the texture's maxUV using a uniform because that way we dont have to update our vertex buffer each frame.
	FFGLTexCoords maxCoords = GetMaxGLTexCoords( *pGL->inputTextures[ 0 ] );
	shader.Set( "MaxUV", maxCoords.s, maxCoords.t );

	//This takes care of sending all the parameter that the plugin registered to the shader.
	sendParams( shader );

	quad.Draw();
	
	//FFGL requires us to leave the context in a default state on return
	resetOpenGLState();
	return FF_SUCCESS;
}
FFResult AddSubtract::DeInitGL()
{
	shader.FreeGLResources();
	quad.Release();

	return FF_SUCCESS;
}
